
# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  socket
from    os.path   import  abspath, isdir, dirname, basename, isfile
from    os.path   import  join                                       as  pjoin
from    os        import  listdir                                    as  os_listdir
from    os        import  mkdir
from    sys       import  argv
from    datetime  import  datetime
from    copy      import  deepcopy

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from core.external      import  tstamp, lmap, lfilter
from core.io_access_ml  import  FPrint_ML, output_dir_ml, Copy_Zipped
from core.gather_data   import  All_Datasets
from core.ml_runner     import  ML_Runner

# ------------------------------------------------------------------------
#                  Basic Functions
# ------------------------------------------------------------------------
listdir_full          =  lambda x: sorted(pjoin(x,y) for y in os_listdir(x))
listdir_full_files    =  lambda x: lfilter(isfile, listdir_full(x))
listdir_full_folders  =  lambda x: lfilter(isdir , listdir_full(x))

def splitpath(x):
    A = []
    iters, Nmax  =  0, max(len(x)*10,int(1e4))
    while basename(x):
        A.append(basename(x))
        x      = dirname(x)
        iters += 1
        assert iters < Nmax
    return A[::-1]

def pop1(A):
    x, = A
    return x

# ------------------------------------------------------------------------
#                  Data Folders
# ------------------------------------------------------------------------

folders_rand  =  [ 'free_Ret_180_Pr_1_f1050_r00_ise_002329_003379_20220822_051126',
                   'free_Ret_180_Pr_1_f1050_r01_ise_001910_002960_20220822_052556',
                   'free_Ret_180_Pr_1_f1050_r02_ise_001897_002947_20220822_054021',
                   'free_Ret_180_Pr_1_f1050_r03_ise_001920_002970_20220822_055445',
                   'free_Ret_180_Pr_1_f1050_r04_ise_001893_002943_20220822_060958',
                   'free_Ret_180_Pr_1_f1050_r05_ise_001886_002936_20220822_062453',
                   'free_Ret_180_Pr_1_f1050_r06_ise_001919_002969_20220822_063915',
                   'free_Ret_180_Pr_1_f1050_r07_ise_001828_002878_20220822_065340',
                   'free_Ret_180_Pr_1_f1050_r08_ise_002140_003190_20220822_070809',
                   'free_Ret_180_Pr_1_f1050_r09_ise_002039_003089_20220822_072228',
                   'free_Ret_180_Pr_1_f1050_r10_ise_002035_003085_20220822_073650',
                   'free_Ret_180_Pr_1_f1050_r11_ise_002006_003056_20220822_075150',
                   'free_Ret_180_Pr_1_f1050_r12_ise_002018_003068_20220822_080639',
                   'free_Ret_180_Pr_1_f1050_r13_ise_001952_003002_20220822_082107',
                   'free_Ret_180_Pr_1_f1050_r14_ise_001838_002888_20220822_083541',
                   'free_Ret_180_Pr_1_f1050_r15_ise_001837_002887_20220822_085002',
                   'free_Ret_180_Pr_1_f1050_r16_ise_002461_003511_20220822_090426',
                   'free_Ret_180_Pr_1_f1050_r17_ise_002304_003354_20220822_091847',
                   'free_Ret_180_Pr_1_f1050_r18_ise_002284_003334_20220822_093318',
                   'free_Ret_180_Pr_1_f1050_r19_ise_002230_003280_20220822_094819',
                   'free_Ret_180_Pr_1_f1050_r20_ise_001741_002791_20220822_100323',
                   'free_Ret_180_Pr_1_f1050_r21_ise_001727_002777_20220822_101744',
                   'free_Ret_180_Pr_1_f1050_r22_ise_001748_002798_20220822_103212',
                   'free_Ret_180_Pr_1_f1050_r23_ise_001727_002777_20220822_104634',
                    'free_Ret_180_Pr_1_f550_r24_ise_000902_001452_20220822_110101',
                    'free_Ret_180_Pr_1_f550_r25_ise_000878_001428_20220822_111704',
                    'free_Ret_180_Pr_1_f550_r26_ise_000902_001452_20220822_113245',
                    'free_Ret_180_Pr_1_f550_r27_ise_000904_001454_20220822_114819',
                    'free_Ret_180_Pr_1_f550_r28_ise_000878_001428_20220822_120342',
                    'free_Ret_180_Pr_1_f550_r29_ise_000891_001441_20220822_122033',
                    'free_Ret_180_Pr_1_f550_r30_ise_000909_001459_20220822_123604',
                    'free_Ret_180_Pr_1_f550_r31_ise_000894_001444_20220822_125148',
                    'free_Ret_180_Pr_1_f550_r32_ise_000905_001455_20220822_130800',
                    'free_Ret_180_Pr_1_f550_r33_ise_001016_001566_20220822_132332',
                    'free_Ret_180_Pr_1_f550_r34_ise_000997_001547_20220822_133916',
                    'free_Ret_180_Pr_1_f550_r35_ise_001017_001567_20220822_135446',
                    'free_Ret_180_Pr_1_f550_r36_ise_001009_001559_20220906_111112',
                    'free_Ret_180_Pr_1_f550_r37_ise_001031_001581_20220906_112724',
                    'free_Ret_180_Pr_1_f550_r38_ise_001014_001564_20220906_114310',
                    'free_Ret_180_Pr_1_f550_r39_ise_001008_001558_20220906_115852']


folders_sel = ['free_Ret_180_Pr_1_f550_sel01_ise_002516_003066_20220821_200941',
               'free_Ret_180_Pr_1_f550_sel02_ise_002535_003085_20220821_201833',
               'free_Ret_180_Pr_1_f550_sel03_ise_002327_002877_20220821_202859',
               'free_Ret_180_Pr_1_f550_sel04_ise_002154_002704_20220821_204405',
               'free_Ret_180_Pr_1_f550_sel05_ise_001327_001877_20220821_205941',
               'free_Ret_180_Pr_1_f550_sel06_ise_001331_001881_20220821_211457',
               'free_Ret_180_Pr_1_f550_sel07_ise_001195_001745_20220821_212921',
               'free_Ret_180_Pr_1_f550_sel08_ise_001253_001803_20220821_214431',
               'free_Ret_180_Pr_1_f550_sel09_ise_001097_001647_20220821_215937',
               'free_Ret_180_Pr_1_f550_sel10_ise_001081_001631_20220821_221448',
               'free_Ret_180_Pr_1_f550_sel11_ise_001072_001622_20220821_222910',
               'free_Ret_180_Pr_1_f550_sel12_ise_001057_001607_20220821_224327',
               'free_Ret_180_Pr_1_f550_sel13_ise_001055_001605_20220821_225838',
               'free_Ret_180_Pr_1_f550_sel14_ise_001088_001638_20220821_231343',
               'free_Ret_180_Pr_1_f550_sel15_ise_001051_001601_20220821_232851',
               'free_Ret_180_Pr_1_f550_sel16_ise_001096_001646_20220821_234314',
               'free_Ret_180_Pr_1_f550_sel17_ise_001050_001600_20220821_235823',
               'free_Ret_180_Pr_1_f550_sel18_ise_001073_001623_20220822_001331',
               'free_Ret_180_Pr_1_f550_sel19_ise_001053_001603_20220822_002835',
               'free_Ret_180_Pr_1_f550_sel20_ise_001056_001606_20220822_004333',
               'free_Ret_180_Pr_1_f550_sel21_ise_001067_001617_20220822_005753',
               'free_Ret_180_Pr_1_f550_sel22_ise_001093_001643_20220822_011218',
               'free_Ret_180_Pr_1_f550_sel23_ise_001115_001665_20220822_012737',
               'free_Ret_180_Pr_1_f550_sel24_ise_001090_001640_20220822_014248',
               'free_Ret_180_Pr_1_f550_sel25_ise_000903_001453_20220822_015748',
               'free_Ret_180_Pr_1_f550_sel26_ise_000885_001435_20220822_021309',
               'free_Ret_180_Pr_1_f550_sel27_ise_000885_001435_20220822_022917',
               'free_Ret_180_Pr_1_f550_sel28_ise_000897_001447_20220822_024523',
               'free_Ret_180_Pr_1_f550_sel29_ise_000898_001448_20220822_030131',
               'free_Ret_180_Pr_1_f550_sel30_ise_000896_001446_20220822_031754',
               'free_Ret_180_Pr_1_f550_sel31_ise_000896_001446_20220822_033428',
               'free_Ret_180_Pr_1_f550_sel32_ise_000903_001453_20220822_035041',
               'free_Ret_180_Pr_1_f550_sel33_ise_000937_001487_20220822_040655',
               'free_Ret_180_Pr_1_f550_sel34_ise_000983_001533_20220822_042241',
               'free_Ret_180_Pr_1_f550_sel35_ise_001018_001568_20220822_043859',
               'free_Ret_180_Pr_1_f550_sel36_ise_001015_001565_20220822_045551',
               'free_Ret_180_Pr_1_f550_sel37_ise_001007_001557_20220906_121435',
               'free_Ret_180_Pr_1_f550_sel38_ise_001011_001561_20220906_123011',
               'free_Ret_180_Pr_1_f550_sel39_ise_001015_001565_20220906_124625',
               'free_Ret_180_Pr_1_f550_sel40_ise_001017_001567_20220906_130347']

folders = folders_rand + [None]*(701-len(folders_rand)) + folders_sel

# ------------------------------------------------------------------------
#                  Seen Argv
# ------------------------------------------------------------------------

def get_log_argv(folder):
    fname, = lfilter(lambda x: '.log' in basename(x),
                     listdir_full_files(folder)     )
    result = []
    with open(fname,'r') as f:
        for x in f:
            x = x.rstrip('\n')
            if '    given_argv' == x[:14]:
                result.append(eval(':'.join(x.split(':')[1:])))
    return pop1(result)

get_seen_argv = lambda qmode: lmap(get_log_argv, listdir_full_folders(pjoin(  dirname(dirname(__file__)),
                                                                             'data'                     ,
                                                                            f'output_{qmode}'           )))

def parse_argv(A):
    A     = deepcopy(A)
    icuda = pop1(lfilter(lambda i: 'cuda' in A[i], [-2,-3]))
    assert 'cuda' in A[icuda]
    assert 'valid_' == A[3][:6]
    assert 'test_'  == A[4][:5]
    A[3]  = 'jointvt_'+'_'.join(map(str,sorted(map(int, A[3].lstrip('valid_').lstrip('test_').split('_') +
                                                        A[4].lstrip('valid_').lstrip('test_').split('_') ))))
    A[4]     = None
    A[icuda] = None
    return A

# ------------------------------------------------------------------------
#                  Run Function
# ------------------------------------------------------------------------

def main_runner(run_tag, qmode, inds_train, inds_valid, inds_test, Cseq, dseq, run_id, device, register_run, weight_decay_w, given_argv):
    
    if parse_argv(given_argv) in lmap(parse_argv, get_seen_argv(qmode)): return
    
    register_run  =  bool(register_run)
    SEED          =  int(datetime.now().strftime("%Y%m%d%H%M%S"))
    
    assert not (set(inds_train) & set(inds_valid))
    assert not (set(inds_train) & set(inds_test ))
    assert not (set(inds_valid) & set(inds_test ))
    
    # -------------------------------------------
    #                  Arguments
    # -------------------------------------------
    
    # arguments (build datasets) 
    args_dsets  =  { 'side'       :  'bottom'                     ,
                     'qmode'      :  qmode                        ,
                     'folders'    :  folders                      , 
                     'inds_train' :  inds_train                   ,
                     'inds_test'  :  inds_test                    ,
                     'inds_valid' :  inds_valid                   ,
                     'as_float'   :  True                         ,
                   }
    
    # arguments (cnn) 
    args_net   =  {'Cseq' : Cseq,
                   'dseq' : dseq}
    
    # arguments (ML runner) 
    args_ML     =  { 'qmode'                :  qmode                                           ,
                     'weight_decay_w'       :  weight_decay_w                                  ,
                     'optim_w_type'         : 'adam'                                           ,
                     'lr_w'                 :  0.001                                           ,
                     'lr_min_w'             :  1e-05                                           ,
                     'Epochs'               :  10001                                           ,
                     'args_net'             :  args_net                                        ,
                     'betas_w'              :  (0.9, 0.999)                                    ,
                     'eps'                  :  1e-7                                            ,
                     'loss_function'        :  'lambda delta: torch.abs(delta).sum()'          , #+ 3*torch.abs(delta.sum())'          , # torch.mean(torch.abs(delta))',# 'lambda delta: torch.mean((delta)**2)' 'lambda delta: torch.mean((delta)**2)'
                     'loss_type'            :  'l1'                                            ,
                     'net_type'             :  'alt2_orig'                                     , # alt2_bigwindow alt2_super2
                     'batch_size'           :  None                                            ,
                     'min_saver_dt'         :  60.                                             ,
                     'breaker'              :  None                                            ,
                     'epoch_save_freq'      :  500                                             ,
                     'epoch_n_saves'        :  None                                            ,
                     'epoch_predict'        :  10                                              ,
                     'epoch_report'         :  10                                              ,
                     'SEED'                 :  SEED                                            ,
                     'register_run'         :  register_run                                    ,
                     'given_argv'           :  given_argv                                      ,
                   }
    
    args_ML['seeder'] = { 'epochs_per_seed':  2000                     ,
                          'n_seeds_try'    :  6 if register_run else 10,
                          'epochs_avg'     :  100                      ,
                          'metric_name'    :  '|diff|/|total|'         ,
                          'use_dsets'      :  ['train','valid'] if register_run else ['train'],
                        }
    
    args_ML['Epochs'] += args_ML['seeder']['epochs_per_seed']*(args_ML['seeder']['n_seeds_try']-1)
    
    assert (args_ML['seeder']['epochs_per_seed'] % args_ML['epoch_save_freq']) == 0
    
    # -------------------------------------------
    #                  Outout Folder
    # -------------------------------------------
    assert type(run_id) == str
    if args_ML['weight_decay_w']:
        tag_wdl2 = "_wdl2_" + f"{(args_ML['weight_decay_w']):.2e}".replace('+','')
    else:
        tag_wdl2 = ''
    subfolder      =  tstamp() + f"_{run_tag}_{args_ML['loss_type']}_{args_ML['qmode']}_net_{args_ML['net_type']}{tag_wdl2}_runid_{run_id}"
    
    output_folder  =  pjoin(  output_dir_ml(qmode)  ,
                              subfolder             )
    
    if not isdir(output_dir_ml(qmode)):
        mkdir(   output_dir_ml(qmode))
    
    assert not isdir(output_folder)
    mkdir(      output_folder         )
    mkdir(pjoin(output_folder,'saved'))
    
    # -------------------------------------------
    #                  Copy Code
    # -------------------------------------------
    
    Copy_Zipped(  dirname(abspath(__file__))  ,
                  output_folder               )
    
    # -------------------------------------------
    #                  Logger
    # -------------------------------------------
    
    # logger
    fprint     =  FPrint_ML( pjoin( output_folder, subfolder + '.log') )
    
    # -------------------------------------------------------------
    #                  Datasets (train/valid/test)
    # -------------------------------------------------------------
    
    # build datasets
    get_all_data   =  lambda: All_Datasets(args_dsets, fprint)
    
    # -------------------------------------------------------------
    #                 ML Runner
    # -------------------------------------------------------------
    
    # ml runner
    ml_runner  =  ML_Runner(get_all_data, args_ML, fprint, output_folder, abspath(__file__), device)
    
    # make predictions
    ml_runner.global_predict()
    
# ------------------------------------------------------------------------
#                  Fetch Inds from Argv
# ------------------------------------------------------------------------

def fetch_inds(A, tag):
    A = A.replace(' ','')
    assert A[:(len(tag)+1)]  ==  (tag + '_')
    A = A.split('_')[1:]
    return [int(x.lstrip('0') or '0') for x in A]

def parse_str(A,s):
    A = A.split('_')
    assert A[0] == s
    A = A[1:]
    for i,x in enumerate(A):
        if 'x' in x:
            A[i] = lmap(int,x.split('x'))
        else:
            A[i] = int(x)
    return A

# ------------------------------------------------------------------------
#                  Direct Runner
# ------------------------------------------------------------------------

if __name__ == '__main__':
    
    run_tag, qmode                     =  lmap( lambda s: s.replace(' ','') ,
                                                argv[1:3]                   )
    
    inds_train, inds_valid, inds_test  =  lmap(  lambda x: fetch_inds(x[0],x[1])          , 
                                                 zip(argv[3:6],['train','valid','test'])  )
    # os_system(f'{python_path} {main_runner} {run_tag} {qmode} train_{group(inds_train)} valid_{group(inds_valid)} test_{group(inds_test)} Cseq_{str_Cseq} dseq_{str_dseq} {run_id}')
    
    Cseq           =  parse_str(argv[6],'Cseq')
    dseq           =  parse_str(argv[7],'dseq')
    run_id         =       argv[ 8]
    device         =       argv[ 9]
    register_run   =  eval(argv[10])
    weight_decay_w =  eval(argv[11])
    given_argv     =  argv[1:]
    main_runner(run_tag, qmode, inds_train, inds_valid, inds_test, Cseq, dseq, run_id, device, register_run, weight_decay_w, given_argv)